﻿/* 
 *  Author:  Flavio Ribeiro (fr@lps.usp.br)
 *  Date:    January 15, 2011
 *  Version: 0.1
 *
 *  Copyright (c) 2011 Flavio Ribeiro
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.  
 */

package 
{	
	import flash.display.*;
	import flash.events.*;
	import flash.text.*;
	import flash.utils.*;
	import flash.geom.Rectangle;
	import flash.geom.Matrix;
	import flash.net.*;
	import flash.system.*;
	import fl.transitions.*;
 	import fl.transitions.easing.None;

	import utils.sprintf;

	Array.prototype.max = function()
	{
		var max = this[0];
		var len = this.length;
		for (var i = 1; i < len; i++)
			if (this[i] > max)
				max = this[i];
		return max;
	}

	public class FCalibrationPlayer extends Sprite
	{
		protected static const BUTTON_WIDTH:Number  = 30;
		protected static const BUTTON_HEIGHT:Number = 26;

		protected static const PLAYING :String = "playing";
		protected static const STOPPED :String = "stopped";
		protected static const PAUSED  :String = "paused";
		protected static const COMPLETE:String = "complete";
		protected static const ERROR   :String = "error";

		protected var _queryString:QueryString = new QueryString();
		protected var _state:String = STOPPED;

		protected var _canvas:Sprite = new Sprite();
		protected var _videoOverlay:Sprite = new Sprite();
		protected var _previewImage:Loader = new Loader();

		private var _bigPlayButton:bigPlayButton = new bigPlayButton();
		private var _bigReplayButton:bigReplayButton = new bigReplayButton();
		private var _errorMessage:TextField = new TextField();

		private var _controlBar:Sprite  = new Sprite();
		private var _playTime:TextField = new TextField();
		private var _progressBar:Sprite = new Sprite();
		private var _progressBarWidth:int;

		protected var _configXML:XML;
		protected var _videoId:String;
		protected var _imageURL:String;

		protected var _replayEnabled:Boolean;
		
		private var _frameCounter:uint = 0;
		private var _coordinates:Array = [];
		private var _timestamp:Array = [];
		private var _elapsedIndex:uint = 0;
		private var _elapsedTime:Number = 0;
		private var _totalTime:Number = 0;
		private var _transition:String;
		
		protected var _ballX:Number;
		protected var _ballY:Number;

		public function FCalibrationPlayer()
		{			
			this.parseArguments();
			this.attachWidgets();
			this.drawWidgets();
			this.loadConfigXML();
		}

		private function initCoordinates():void
		{
			var x:Array = [-1, 1, 1, -1, -1, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, -1, 0, 0, 1, 1, 0, -1, 0, 1, 0, 1, 0, 1, 0, 1, 0, -1, -1, -1, 0, 1, 0, -1, 0, 0, -1, 1, 0, 1, 1, 0, -1, 1, 0, 1, -1, -1, -1, -1, -1, 0, 0, -1, -1, 1, 1, 0, -1, -1, -1, -1, -1, 1, 1, 0, 1, -1, 1, 0, 0, 0, -1, 1, 0, 0, 0, -1, 1, -1, -1, 1, 1, -1, 0, -1, -1, 1, 0, 1, -1, 0, 1, 0, 0, 0];
			var y:Array = [1, 1, 0, -1, 0, -1, 1, 0, 1, -1, 0, -1, 1, -1, 0, -1, 0, 0, 0, 1, 1, -1, 1, 1, -1, -1, -1, 0, -1, -1, -1, -1, -1, 1, 0, -1, -1, 0, -1, 0, -1, 1, -1, -1, 0, 0, 1, 1, 0, 0, -1, 1, 1, -1, 1, 1, 0, -1, -1, 0, 1, -1, 1, 0, 0, -1, 0, -1, -1, 1, -1, -1, 0, -1, 1, -1, -1, -1, 1, 0, 1, 0, -1, 1, -1, -1, -1, 0, -1, -1, -1, -1, -1, 1, 1, 0, 0, -1, 1, 1];
			var z:Array = [1, 9, 8, 4, 2, 4, 4, 7, 4, 9, 9, 9, 4, 3, 8, 2, 2, 5, 9, 4, 2, 1, 9, 6, 4, 6, 8, 4, 7, 2, 9, 7, 2, 4, 6, 9, 5, 6, 7, 9, 2, 4, 3, 7, 9, 5, 8, 3, 5, 4, 9, 7, 8, 1, 3, 7, 2, 1, 4, 9, 3, 6, 4, 1, 5, 8, 3, 1, 3, 6, 3, 6, 1, 7, 4, 1, 4, 7, 9, 9, 9, 9, 8, 8, 5, 4, 2, 6, 8, 3, 8, 1, 4, 3, 2, 6, 3, 9, 5, 7];
			var x0 = _canvas.width/2;
			var y0 = _canvas.height/2;
			var i:uint;
			
			x = x.slice(0,80);
			y = y.slice(0,80);
			z = z.slice(0,80);
			
			_coordinates = new Array();
			for (i = 0 ; i < x.length; i++)
				_coordinates[i] = [ 100*x[i] + x0, 60*y[i] + y0 ];
			for (i = 20; i < x.length; i++)
				_coordinates[i] = [ 150*x[i] + x0, 80*y[i] + y0 ];
			for (i = 40; i < x.length; i++)
				_coordinates[i] = [ 200*x[i] + x0, 100*y[i] + y0 ];
			for (i = 60; i < x.length; i++)
				_coordinates[i] = [ 250*x[i] + x0, 120*y[i] + y0 ];

			if (_transition == "translate")
			{
				var dx:Number;
				var dy:Number;
				
				_timestamp[0] = 0;
				for (i = 1; i < x.length; i++)
				{
					dx = _coordinates[i][0] - _coordinates[i-1][0];
					dy = _coordinates[i][1] - _coordinates[i-1][1];
					_timestamp[i] = _timestamp[i-1] + Math.sqrt(dx*dx + dy*dy)/130;
				}
			}
			else
			{
				for (i = 0; i < x.length; i++)
					_timestamp[i] = 1.5*i + z[i]/10;
			}

			_totalTime = _timestamp.max();
		}

		private function isLocal():Boolean
		{
			return Capabilities.playerType == "Standalone" || Capabilities.playerType == "External" || Security.sandboxType != Security.REMOTE;
		}

		public function setImageURL(url:String):void
		{
			if (!_imageURL)
			{
				// the ?time= trick prevents caching when running as a browser plug-in
				if (!url || url == "" || isLocal())
					_imageURL = url;
				else
					_imageURL = url + "?time=" + new Date().getTime();
			}
		}

		public function renderState(state:String):void
		{
			if (_state == ERROR)
				return;

			_bigPlayButton.visible = false;
			_bigReplayButton.visible = false;
			_previewImage.visible = false;
			_errorMessage.visible = false;
			_state = state;

			switch(state)
			{
				case STOPPED:
					_bigPlayButton.visible = true;
					_previewImage.visible = true;
					break;

				case PLAYING:
				case PAUSED:
					break;

				case COMPLETE:
					_bigReplayButton.visible = _replayEnabled;
					break;

				case ERROR:
					_errorMessage.visible = true;
					break;

				default:
					break;
			}
		}

		public function printError(message:String):void
		{
			this.renderState(ERROR);
			pauseVideo(null);

			while (_videoOverlay.numChildren > 0)
				_videoOverlay.removeChildAt(0);
			_videoOverlay.graphics.beginFill(0x000000, 0.5);
			_videoOverlay.graphics.drawRect(0, 0, stage.stageWidth, stage.stageHeight);
			_videoOverlay.graphics.endFill();

			_errorMessage.htmlText = message;
			_errorMessage.x = stage.stageWidth/2 - _errorMessage.width/2;
			_errorMessage.y = stage.stageHeight/2 - _errorMessage.height/2;
		}

		public function playVideo(e:MouseEvent)
		{	
			if (_state != PLAYING)
				this.renderState(PLAYING);
		}

		public function pauseVideo(e:MouseEvent)
		{
			if (_state != PAUSED)
				this.renderState(PAUSED);
		}

		public function videoComplete():void
		{
			this.renderState(COMPLETE);
		}

		public function replayVideo(e:MouseEvent)
		{
			_frameCounter = 0;
			playVideo(null);
		}

		protected function processConfigXML(e:Event):void
		{
			XML.ignoreComments = true;
			XML.ignoreProcessingInstructions = true;
			XML.ignoreWhitespace = true;

			try
			{
				_configXML  = new XML(e.target.data);
				_transition = _configXML.video.(@id == _videoId).transition.toString();
				setImageURL(_configXML.video.(@id == _videoId).imageURL.toString());
			}
			catch(e:Error)
			{
				_configXML = null;
				setImageURL(null);
			}

			initCoordinates();
			startPlayer();
		}

		private function startPlayer():void
		{
			if (!this.hasErrors())
			{
				this.addEventCallbacks();
				this.loadPreviewImage();
				this.renderState(STOPPED);
			}
		}

		private function hasErrors():Boolean
		{
			if (stage.stageWidth < 640 || stage.stageHeight < 386)
			{
				printError("Error: please reset your browser's zoom settings and reload this page.");
				return true;
			}

			return false;
		}

		private function parseArguments():void
		{
			var parameters = LoaderInfo(this.root.loaderInfo).parameters;

			_replayEnabled = parameters.replay == "off" ? false : true;

			_videoId    = _queryString.parameters.videoId;
			if (!_videoId)
				_videoId = "";
		}

		private function loadConfigXML():void
		{
			var xmlLoader:URLLoader = new URLLoader();
			var xmlRequest:URLRequest = new URLRequest();

			if (isLocal())
				xmlRequest.url = "videos.xml";
			else
				xmlRequest.url = "videos.xml?time=" + new Date().getTime();

			xmlLoader.addEventListener(Event.COMPLETE, processConfigXML);
			xmlLoader.addEventListener(IOErrorEvent.IO_ERROR, loadConfigXMLError);
			xmlLoader.load(xmlRequest);
		}

		private function attachWidgets():void
		{
			addChild(_canvas);
			addChild(_previewImage);
			addChild(_videoOverlay);

			addChild(_errorMessage);
			addChild(_bigPlayButton);
			addChild(_bigReplayButton);

			addChild(_controlBar);
			_controlBar.addChild(_playTime);
			_controlBar.addChild(_progressBar);
		}

		private function drawWidgets():void
		{
			var gradientMatrix:Matrix = new Matrix();
			gradientMatrix.createGradientBox(BUTTON_WIDTH, BUTTON_HEIGHT, Math.PI/2);

			stage.align     = StageAlign.TOP_LEFT;
			stage.scaleMode = StageScaleMode.NO_SCALE;

			_canvas.x = 0;
			_canvas.y = 0;			
			_canvas.graphics.clear();
			_canvas.graphics.beginFill(0x000000);
			_canvas.graphics.drawRect(0, 0, stage.stageWidth, stage.stageHeight - BUTTON_HEIGHT);
			_canvas.graphics.endFill();

			_errorMessage.defaultTextFormat = new TextFormat("_sans", 14, 0xffffff);
			_errorMessage.autoSize = TextFieldAutoSize.CENTER;

			// Set up Control Bar
			_controlBar.graphics.clear();
			_controlBar.graphics.beginGradientFill(GradientType.LINEAR, [0xFFFFFF, 0xCCCCCC], [1, 1], [0x00, 0xFF], gradientMatrix);
			_controlBar.graphics.drawRect(0, 0, stage.stageWidth, BUTTON_HEIGHT);
			_controlBar.graphics.endFill();

			_controlBar.x = 0;
			_controlBar.y = stage.stageHeight - BUTTON_HEIGHT;

			_playTime.defaultTextFormat = new TextFormat("_sans", 10, 0x000000, false);
			_playTime.autoSize = TextFieldAutoSize.LEFT;
			_playTime.text = "00:00 / 00:00";
			_playTime.x = 4;
			_playTime.y = 5;

			_progressBarWidth = stage.stageWidth - _playTime.x - _playTime.width - 4;
			_progressBar.graphics.clear();
			_progressBar.graphics.beginFill(0x000000, .5);
			_progressBar.graphics.drawRect(0, 0, _progressBarWidth, _controlBar.height);
			_progressBar.graphics.endFill();
			_progressBar.x = _playTime.x + _playTime.width + 4;
			_progressBar.y = 0;
			_progressBar.buttonMode = false;

			_bigPlayButton.x = (stage.stageWidth - _bigPlayButton.width)/2;
			_bigPlayButton.y = (stage.stageHeight - BUTTON_HEIGHT - _bigPlayButton.height)/2;

			_bigReplayButton.x = (stage.stageWidth - _bigReplayButton.width)/2;
			_bigReplayButton.y = (stage.stageHeight - BUTTON_HEIGHT - _bigReplayButton.height)/2;

			this.renderState(_state);
		}

		private function addEventCallbacks():void
		{
			addEventListener(Event.ENTER_FRAME, updateProgress);
			addEventListener(Event.ENTER_FRAME, renderFrame);

			_previewImage.contentLoaderInfo.addEventListener(Event.COMPLETE, loadPreviewImageComplete);
			_previewImage.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, loadPreviewImageError);
			_bigReplayButton.addEventListener(MouseEvent.CLICK, replayVideo);
			_bigPlayButton.addEventListener(MouseEvent.CLICK, playVideo);			
		}

		private function loadPreviewImage():void
		{
			// Check for preview image and setup display
			if (_imageURL && _imageURL != "")
				_previewImage.load(new URLRequest(_imageURL));
		}

		//
		// Event Handlers
		//
		
		private function loadConfigXMLError(e:IOErrorEvent):void
		{
			printError("Error: cannot load configuration XML file.");
		}
		
		private function loadPreviewImageError(e:IOErrorEvent):void
		{
			printError(sprintf("Error: cannot load preview image %s.", _imageURL));
		}

		private function loadPreviewImageComplete(e:Event):void
		{
			_previewImage.x = (stage.stageWidth - _previewImage.width)/2;
			_previewImage.y = (stage.stageHeight - BUTTON_HEIGHT - _previewImage.height)/2;
		}

		private function updateProgress(e:Event)
		{
			var barWidth:int;
			_elapsedTime = _frameCounter / 24;

			_progressBar.graphics.clear();
			_progressBar.graphics.beginFill(0x000000, .5);
			_progressBar.graphics.drawRect(0, 0, _progressBarWidth, _controlBar.height);
			_progressBar.graphics.endFill();

			barWidth = _progressBar.width;
			_progressBar.graphics.beginFill(0xFF0000, .2);
			_progressBar.graphics.drawRect(0, 0, barWidth, _progressBar.height);
			_progressBar.graphics.endFill();

			barWidth = Math.round((_elapsedTime / _totalTime) * _progressBar.width);
			_progressBar.graphics.beginFill(0xFF0000, .5);
			_progressBar.graphics.drawRect(0, 0, barWidth, _progressBar.height);
			_progressBar.graphics.endFill();
				
			_playTime.text = sprintf("%02d:%02d / %02d:%02d", Math.floor(_elapsedTime / 60), Math.floor(_elapsedTime % 60),
						   									  Math.floor(_totalTime / 60), Math.floor(_totalTime % 60));			
		}
		
		private function renderFrame(e:Event)
		{
			if (_state == PLAYING)
			{
				while (_elapsedIndex < (_coordinates.length - 1) && _timestamp[_elapsedIndex] < _elapsedTime)
					_elapsedIndex++;
				while (_elapsedIndex > 0 && _timestamp[_elapsedIndex] > _elapsedTime)
					_elapsedIndex--;
									
				if (_elapsedIndex == (_coordinates.length - 1))
				{
					videoComplete();
					return;
				}

				var h:Number;
				if (_transition == "translate")
					h = (_elapsedTime - _timestamp[_elapsedIndex]) / (_timestamp[_elapsedIndex+1] - _timestamp[_elapsedIndex]);
				else
					h = 0;

				_ballX = (1-h)*_coordinates[_elapsedIndex][0] + h*_coordinates[_elapsedIndex+1][0];
				_ballY = (1-h)*_coordinates[_elapsedIndex][1] + h*_coordinates[_elapsedIndex+1][1];
				
				_canvas.graphics.clear();
				_canvas.graphics.beginFill(0x000000);
				_canvas.graphics.drawRect(0, 0, stage.stageWidth, stage.stageHeight - BUTTON_HEIGHT);
				_canvas.graphics.endFill();
				_canvas.graphics.beginFill(0xFF0000);
				_canvas.graphics.drawCircle(_ballX, _ballY, 10);
				_canvas.graphics.endFill();

				_frameCounter++;
			}
		}
	}
}
