﻿/* 
 *  Author:  Flavio Ribeiro (fr@lps.usp.br)
 *  Date:    January 15, 2011
 *  Version: 0.1
 *
 *  Copyright (c) 2011 Flavio Ribeiro
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.  
 */

package 
{	
	import flash.display.*;
	import flash.events.*;
	import flash.text.*;
	import flash.utils.*;
	import flash.geom.Rectangle;
	import flash.geom.Matrix;
	import flash.net.*;
	import flash.system.*;
	import fl.video.*;
	import fl.transitions.*;
 	import fl.transitions.easing.None;

	import utils.sprintf;

	public class FVideoPlayer extends Sprite
	{
		private static const BUTTON_WIDTH:Number  = 30;
		private static const BUTTON_HEIGHT:Number = 26;

		protected static const PLAYING :String = "playing";
		protected static const STOPPED :String = "stopped";
		protected static const PAUSED  :String = "paused";
		protected static const COMPLETE:String = "complete";
		protected static const ERROR   :String = "error";

		protected var _queryString:QueryString = new QueryString();
		protected var _state:String = STOPPED;

		protected var _videoPlayback:FLVPlayback = new FLVPlayback();
		protected var _videoOverlay:Sprite = new Sprite();
		protected var _previewImage:Loader = new Loader();

		private var _loadingAnim:loadingAnim = new loadingAnim();
		private var _loadingText:TextField = new TextField();
		private var _bigPlayButton:bigPlayButton = new bigPlayButton();
		private var _bigReplayButton:bigReplayButton = new bigReplayButton();
		private var _errorMessage:TextField = new TextField();

		private var _controlBar:Sprite         = new Sprite();
		private var _playButton:playButton     = new playButton();
		private var _pauseButton:pauseButton   = new pauseButton();
		private var _volumeButton:volumeButton = new volumeButton();
		private var _playTime:TextField        = new TextField();
		private var _progressBar:Sprite        = new Sprite();
		private var _progressBarWidth:int;
		private var _startTime:Number;
		private var _lastPlaybackTime:Number = 0;
		private var _lastBytesLoaded:int     = 0;
		private var _lastBytesTotal:int      = 0;

		private var _volumePanel:MovieClip = new MovieClip();
		private var _volumePost:Sprite = new Sprite();
		private var _volumeSlider:Sprite = new Sprite();
		protected var _volumeMouseOver:Boolean;
		protected var _volumeMouseDown:Boolean;

		protected var _configXML:XML;
		protected var _videoId:String;
		protected var _videoURL:String;
		protected var _imageURL:String;

		protected var _volume:Number;
		protected var _mute:Boolean;
		protected var _controls:Boolean;
		protected var _replayEnabled:Boolean;

		public function FVideoPlayer()
		{
			this.parseArguments();
			this.attachWidgets();
			this.drawWidgets();
			this.setVolume();
			this.loadConfigXML();
		}

		private function isLocal():Boolean
		{
			return Capabilities.playerType == "Standalone" || Capabilities.playerType == "External" || Security.sandboxType != Security.REMOTE;
		}

		public function setVideoURL(url:String):void
		{
			if (!_videoURL)
			{
				// the ?time= trick prevents caching when running as a browser plug-in
				if (!url || url == "" || isLocal())
					_videoURL = url;
				else
					_videoURL = url + "?time=" + new Date().getTime();
			}
		}

		public function setImageURL(url:String):void
		{
			if (!_imageURL)
			{
				// the ?time= trick prevents caching when running as a browser plug-in
				if (!url || url == "" || isLocal())
					_imageURL = url;
				else
					_imageURL = url + "?time=" + new Date().getTime();
			}
		}

		public function setVolume():void
		{
			if (_mute)
			{
				setVolumeSlider(0);
				_videoPlayback.volume = 0;
			}
			else
			{
				setVolumeSlider(_volume);
				_videoPlayback.volume = _volume;
			}
		}

		public function renderState(state:String):void
		{
			if (_state == ERROR)
				return;

			_loadingAnim.visible = false;
			_loadingText.visible = false;
			_bigPlayButton.visible = false;
			_bigReplayButton.visible = false;
			_previewImage.visible = false;
			_errorMessage.visible = false;
			_state = state;

			switch(state)
			{
				case STOPPED:
					_playButton.visible = true;
					_pauseButton.visible = false;

					_bigPlayButton.visible = true;
					_previewImage.visible = true;
					break;

				case PLAYING:
					_playButton.visible = false;
					_pauseButton.visible = true;
					break;

				case PAUSED:
					_playButton.visible = true;
					_pauseButton.visible = false;
					break;

				case COMPLETE:
					_bigReplayButton.visible = _replayEnabled;
					break;

				case ERROR:
					_errorMessage.visible = true;
					break;

				default:
					break;
			}
		}

		public function printError(message:String):void
		{
			this.renderState(ERROR);
			pauseVideo(null);

			while (_videoOverlay.numChildren > 0)
				_videoOverlay.removeChildAt(0);
			_videoOverlay.graphics.beginFill(0x000000, 0.5);
			_videoOverlay.graphics.drawRect(0, 0, stage.stageWidth, stage.stageHeight);
			_videoOverlay.graphics.endFill();

			_errorMessage.htmlText = message;
			_errorMessage.x = _videoPlayback.registrationWidth/2 - _errorMessage.width/2;
			_errorMessage.y = _videoPlayback.registrationHeight/2 - _errorMessage.height/2;
		}

		public function playVideo(e:MouseEvent)
		{	
			if (_state != PLAYING)
			{
				this.renderState(PLAYING);
				if (_videoPlayback.source == null || _videoPlayback.source == "")
					_videoPlayback.load(_videoURL);
				videoStateChange(null);
				playWhenEnoughDownloaded();
			}
		}

		public function pauseVideo(e:MouseEvent)
		{
			if (_state != PAUSED)
			{
				this.renderState(PAUSED);
				_videoPlayback.pause();
			}
		}

		public function videoComplete(e:VideoEvent):void
		{
			this.renderState(COMPLETE);
		}

		public function replayVideo(e:MouseEvent)
		{
			_videoPlayback.playheadTime = 0;
			playVideo(null);
		}

		protected function processConfigXML(e:Event):void
		{
			XML.ignoreComments = true;
			XML.ignoreProcessingInstructions = true;
			XML.ignoreWhitespace = true;

			try
			{
				_configXML = new XML(e.target.data);
				setVideoURL(_configXML.video.(@id == _videoId).videoURL.toString());
				setImageURL(_configXML.video.(@id == _videoId).imageURL.toString());
			}
			catch(e:Error)
			{
				_configXML = null;
				setVideoURL(null);
				setImageURL(null);
			}

			startPlayer();
		}

		private function startPlayer():void
		{
			if (!this.hasErrors())
			{
				this.addEventCallbacks();
				this.loadPreviewImage();
				this.renderState(STOPPED);
				_startTime = getTimer();
				_videoPlayback.load(_videoURL);
			}
		}

		private function hasErrors():Boolean
		{
			if (stage.stageWidth < 640 || stage.stageHeight < 386)
			{
				printError("Error: please reset your browser's zoom settings and reload this page.");
				return true;
			}

			// Display error message if video path is null or incorrect
			if (!_videoURL || _videoURL == "")
			{
				printError(sprintf("Error: no FLV file was specified (videoId = %s).", _videoId));
				return true;
			}

			return false;
		}

		private function parseArguments():void
		{
			var parameters = LoaderInfo(this.root.loaderInfo).parameters;

			_videoId  = parameters.videoId;
			_volume   = parameters.volume == null ? -1 : parameters.volume;
			_mute     = parameters.mute == "on" ? true : false;
			_controls = parameters.controls == "off" ? false : true;
			_replayEnabled = parameters.replay == "off" ? false : true;

			if (!_videoId || _videoId == "")
				_videoId  = _queryString.parameters.videoId;

			if (!_videoId)
				_videoId = "";

			if (!_volume || _volume > 95 || _volume < 0)
				_volume  = 0.95;
			else
				_volume /= 100;
		}

		private function loadConfigXML():void
		{
			var xmlLoader:URLLoader = new URLLoader();
			var xmlRequest:URLRequest = new URLRequest();

			if (isLocal())
				xmlRequest.url = "videos.xml";
			else
				xmlRequest.url = "videos.xml?time=" + new Date().getTime();

			xmlLoader.addEventListener(Event.COMPLETE, processConfigXML);
			xmlLoader.addEventListener(IOErrorEvent.IO_ERROR, loadConfigXMLError);
			xmlLoader.load(xmlRequest);
		}

		private function attachWidgets():void
		{
			addChild(_videoPlayback);
			addChild(_previewImage);
			addChild(_videoOverlay);

			addChild(_errorMessage);
			addChild(_bigPlayButton);
			addChild(_bigReplayButton);
			addChild(_loadingAnim);
			addChild(_loadingText);

			addChild(_volumePanel);
			_volumePanel.addChild(_volumePost);
			_volumePanel.addChild(_volumeSlider);

			addChild(_controlBar);
			if (_controls)
			{
				_controlBar.addChild(_playButton);
				_controlBar.addChild(_pauseButton);
				_controlBar.addChild(_volumeButton);
			}
			_controlBar.addChild(_playTime);
			_controlBar.addChild(_progressBar);
		}

		private function drawWidgets():void
		{
			var gradientMatrix:Matrix = new Matrix();
			gradientMatrix.createGradientBox(BUTTON_WIDTH, BUTTON_HEIGHT, Math.PI/2);

			stage.align     = StageAlign.TOP_LEFT;
			stage.scaleMode = StageScaleMode.NO_SCALE;

			// Setup Video Playback
//			_videoPlayback.getVideoPlayer(_videoPlayback.activeVideoPlayerIndex).smoothing = true;
			_videoPlayback.fullScreenTakeOver = false;
			_videoPlayback.registrationWidth  = stage.stageWidth;
			_videoPlayback.registrationHeight = stage.stageHeight - BUTTON_HEIGHT;
			_videoPlayback.scaleMode = VideoScaleMode.NO_SCALE;
			_videoPlayback.registrationX = 0;
			_videoPlayback.registrationY = 0;

			_errorMessage.defaultTextFormat = new TextFormat("_sans", 14, 0xffffff);
			_errorMessage.autoSize = TextFieldAutoSize.CENTER;

			// Set up Control Bar
			_controlBar.graphics.clear();
			_controlBar.graphics.beginGradientFill(GradientType.LINEAR, [0xFFFFFF, 0xCCCCCC], [1, 1], [0x00, 0xFF], gradientMatrix);
			_controlBar.graphics.drawRect(0, 0, stage.stageWidth, BUTTON_HEIGHT);
			_controlBar.graphics.endFill();

			if (_controls)
			{
				_controlBar.graphics.lineStyle(1, 0xA0A0A0, 1.0);
				_controlBar.graphics.moveTo(BUTTON_WIDTH, 0);
				_controlBar.graphics.lineTo(BUTTON_WIDTH, BUTTON_HEIGHT);
				_controlBar.graphics.moveTo(2*BUTTON_WIDTH, 0);
				_controlBar.graphics.lineTo(2*BUTTON_WIDTH, BUTTON_HEIGHT);
			}

			_controlBar.x = 0;
			_controlBar.y = stage.stageHeight - BUTTON_HEIGHT;

			_playButton.x = 0;
			_playButton.y = 0;

			_pauseButton.x = 0;
			_pauseButton.y = 0;

			_volumeButton.x = BUTTON_WIDTH;
			_volumeButton.y = 0;

			_playTime.defaultTextFormat = new TextFormat("_sans", 10, 0x000000, false);
			_playTime.autoSize = TextFieldAutoSize.LEFT;
			_playTime.text = "00:00 / 00:00";
			_playTime.x = _controls ? (2*BUTTON_WIDTH + 4) : 4;
			_playTime.y = 5;

			_progressBarWidth = stage.stageWidth - _playTime.x - _playTime.width - 4;
			_progressBar.graphics.clear();
			_progressBar.graphics.beginFill(0x000000, .5);
			_progressBar.graphics.drawRect(0, 0, _progressBarWidth, _controlBar.height);
			_progressBar.graphics.endFill();
			_progressBar.x = _playTime.x + _playTime.width + 4;
			_progressBar.y = 0;
			_progressBar.buttonMode = _controls ? true : false;

			_loadingAnim.x = (stage.stageWidth - _loadingAnim.width)/2;
			_loadingAnim.y = (stage.stageHeight - BUTTON_HEIGHT - _loadingAnim.height)/2 - 10;

			_loadingText.defaultTextFormat = new TextFormat("_sans", 14, 0xffffff);
			_loadingText.autoSize = TextFieldAutoSize.CENTER;
			_loadingText.text = "";
			_loadingText.x = stage.stageWidth/2 - _loadingText.width/2;
			_loadingText.y = _loadingAnim.y + _loadingAnim.height + 10;

			_bigPlayButton.x = (stage.stageWidth - _bigPlayButton.width)/2;
			_bigPlayButton.y = (stage.stageHeight - BUTTON_HEIGHT - _bigPlayButton.height)/2;

			_bigReplayButton.x = (stage.stageWidth - _bigReplayButton.width)/2;
			_bigReplayButton.y = (stage.stageHeight - BUTTON_HEIGHT - _bigReplayButton.height)/2;

			gradientMatrix.createGradientBox(BUTTON_WIDTH, 80, Math.PI/2);
			_volumePanel.graphics.clear();
			_volumePanel.graphics.beginGradientFill(GradientType.LINEAR, [0xFFFFFF, 0xCCCCCC], [1, 1], [0x00, 0xFF], gradientMatrix);
			_volumePanel.graphics.drawRect(1, 0, BUTTON_WIDTH-1, 80);
			_volumePanel.x = BUTTON_WIDTH;
			_volumePanel.y = _controlBar.y - _volumePanel.height;
			_volumePanel.visible = false;

			gradientMatrix.createGradientBox(6, _volumePanel.height - 16, 0);
			_volumePost.graphics.clear();
			_volumePost.graphics.beginGradientFill(GradientType.LINEAR, [0x999999, 0xBBBBBB], [1, 1], [0x00, 0xFF], gradientMatrix);
			_volumePost.graphics.drawRoundRect(0, 0, 6, _volumePanel.height - 16, 2, 2);
			_volumePost.graphics.endFill();
			_volumePost.x = Math.round(_volumePanel.width/2 - 3);
			_volumePost.y = 8;
			_volumePost.buttonMode = true;

			gradientMatrix.createGradientBox(16, 8, Math.PI/2);
			_volumeSlider.graphics.clear();
			_volumeSlider.graphics.beginGradientFill(GradientType.LINEAR, [0xFFFFFF, 0xBBBBBB], [1, 1], [0x00, 0xFF], gradientMatrix);
			_volumeSlider.graphics.lineStyle(1, 0x999999);
			_volumeSlider.graphics.drawRect(0, 0, 16, 8);
			_volumeSlider.graphics.endFill();
			_volumeSlider.buttonMode = true;
			_volumeSlider.x = Math.round(_volumePost.x + _volumePost.width/2 - _volumeSlider.width/2);
			_volumeSlider.y = 0;
			
			setVolume();
			this.renderState(_state);
		}

		private function addEventCallbacks():void
		{
			addEventListener(Event.ENTER_FRAME, updateProgress);

			_previewImage.contentLoaderInfo.addEventListener(Event.COMPLETE, loadPreviewImageComplete);
			_previewImage.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, loadPreviewImageError);
			_videoPlayback.addEventListener(VideoEvent.STATE_CHANGE, videoStateChange);
			_videoPlayback.addEventListener(VideoEvent.COMPLETE, videoComplete);
			_videoPlayback.addEventListener(MouseEvent.CLICK, playVideo);
			
			if (_controls)
			{
				_playButton.addEventListener(MouseEvent.CLICK, playVideo);
				_pauseButton.addEventListener(MouseEvent.CLICK, pauseVideo);
				_volumeButton.addEventListener(MouseEvent.CLICK, toggleVolumePanel);
				_progressBar.addEventListener(MouseEvent.CLICK, seekVideo);
	
				_volumePanel.addEventListener(MouseEvent.ROLL_OVER, volumePanelRollOver);
				_volumePanel.addEventListener(MouseEvent.ROLL_OUT, volumePanelRollOut);
				_volumeSlider.addEventListener(MouseEvent.MOUSE_DOWN, volumeSliderMouseDown);
				_volumeSlider.addEventListener(MouseEvent.MOUSE_MOVE, volumeSliderMouseMove);
				_volumePost.addEventListener(MouseEvent.CLICK, volumePostClick);
			}

			_bigReplayButton.addEventListener(MouseEvent.CLICK, replayVideo);
			_bigPlayButton.addEventListener(MouseEvent.CLICK, playVideo);
		}

		private function loadPreviewImage():void
		{
			// Check for preview image and setup display
			if (_imageURL && _imageURL != "")
				_previewImage.load(new URLRequest(_imageURL));
		}

		//
		// Event Handlers
		//
		
		private function loadConfigXMLError(e:IOErrorEvent):void
		{
			printError("Error: cannot load configuration XML file.");
		}
		
		private function loadPreviewImageError(e:IOErrorEvent):void
		{
			printError(sprintf("Error: cannot load preview image %s.", _imageURL));
		}
		
		private function loadPreviewImageComplete(e:Event):void
		{
			_previewImage.x = (stage.stageWidth - _previewImage.width)/2;
			_previewImage.y = (stage.stageHeight - BUTTON_HEIGHT - _previewImage.height)/2;
		}

		private function videoStateChange(e:VideoEvent):void
		{
			switch(_videoPlayback.state)
			{
				case VideoState.CONNECTION_ERROR:
					printError("Error: cannot load video.");
					break;
					
				case VideoState.STOPPED:
				case VideoState.PAUSED:
				case VideoState.BUFFERING:
				case VideoState.LOADING:
				case VideoState.REWINDING:
				case VideoState.SEEKING:
					if (_state == PLAYING)
					{
						if (!_loadingAnim.visible)
						{
							_loadingText.visible = true;
							_loadingAnim.visible = true;
							_loadingAnim.play();
						}
					}
					else
					{
						_loadingText.visible = false;
						_loadingAnim.visible = false;
						_loadingAnim.stop();
					}
					break;
					
				case VideoState.PLAYING:
					if (_state == PLAYING)
					{
						_loadingText.visible = false;
						_loadingAnim.visible = false;
						_loadingAnim.stop();
					}
					break;
					
				default:
					break;
			}
		}

		private function setFullScreen(e:MouseEvent)
		{
			if (stage.displayState == StageDisplayState.NORMAL)
				stage.displayState = StageDisplayState.FULL_SCREEN;
			else if (stage.displayState == StageDisplayState.FULL_SCREEN)
				stage.displayState = StageDisplayState.NORMAL;
		}
		
		private function fullScreenEvent(event:FullScreenEvent):void
		{
			if (stage.displayState == StageDisplayState.NORMAL)
				_controlBar.visible = true;
			else if (stage.displayState == StageDisplayState.FULL_SCREEN)
				_controlBar.visible = false;
			this.drawWidgets();
		}

		private function getBufferingTime():Number
		{
			return (getTimer() - _startTime)/1000;
		}

		private function playWhenEnoughDownloaded():void
		{
			if (isNaN(_videoPlayback.playheadTime) || isNaN(_videoPlayback.totalTime) || getBufferingTime() < 1.0)
			{
				setTimeout(playWhenEnoughDownloaded,1000);
				return;
			}

			var A = 1.5;	// bandwidth safety margin (must be > 1.0)
			var B = _videoPlayback.bytesLoaded / _videoPlayback.bytesTotal * _videoPlayback.totalTime;
			var D = _videoPlayback.totalTime;
			var T = getBufferingTime();

			if (B/T * D > A*(D - B))
			{
				_loadingText.text = "";
				if (_state == PLAYING)
					_videoPlayback.play();
			}
			else
			{
				_loadingText.text = sprintf("Loading... %3d%%", Math.round(100*(A*B)/(A*D - B*D/T)));
				setTimeout(playWhenEnoughDownloaded,1000);
			}
		}

		private function updateProgress(e:Event)
		{
			if (_lastPlaybackTime == _videoPlayback.playheadTime &&
				_lastBytesLoaded  == _videoPlayback.bytesLoaded &&
				_lastBytesTotal  == _videoPlayback.bytesTotal)
				return;

			if (!isNaN(_videoPlayback.playheadTime) && !isNaN(_videoPlayback.totalTime))
			{
				var totalTime:int   = Math.floor(_videoPlayback.totalTime);			
				var elapsedTime:int = Math.floor(_videoPlayback.playheadTime);
				var barWidth:int;

				_progressBar.graphics.clear();
				_progressBar.graphics.beginFill(0x000000, .5);
				_progressBar.graphics.drawRect(0, 0, _progressBarWidth, _controlBar.height);
				_progressBar.graphics.endFill();

				if (_videoPlayback.bytesTotal && _videoPlayback.bytesTotal > 0)
				{
					barWidth = Math.round((_videoPlayback.bytesLoaded / _videoPlayback.bytesTotal) * _progressBar.width);
					_progressBar.graphics.beginFill(0xFF0000, .2);
					_progressBar.graphics.drawRect(0, 0, barWidth, _progressBar.height);
					_progressBar.graphics.endFill();
				}

				barWidth = Math.round((_videoPlayback.playheadTime / _videoPlayback.totalTime) * _progressBar.width);
				_progressBar.graphics.beginFill(0xFF0000, .5);
				_progressBar.graphics.drawRect(0, 0, barWidth, _progressBar.height);
				_progressBar.graphics.endFill();
				
				_playTime.text = sprintf("%02d:%02d / %02d:%02d", Math.floor(elapsedTime / 60), Math.floor(elapsedTime % 60),
							   									  Math.floor(totalTime / 60), Math.floor(totalTime % 60));
				
				// detect if the connection was unexpectedly closed (which makes Flash Player 10 cut the video short)
				if (_videoPlayback.bytesTotal < _lastBytesTotal)
					printError('Connection error while loading video.');
			}
			
			_lastPlaybackTime = _videoPlayback.playheadTime;
			_lastBytesLoaded  = _videoPlayback.bytesLoaded;
			_lastBytesTotal   = _videoPlayback.bytesTotal;
		}

		private function seekVideo(e:MouseEvent)
		{
			if(!isNaN(_videoPlayback.totalTime))
				_videoPlayback.playheadTime = Math.round(_progressBar.mouseX / _progressBar.width * _videoPlayback.totalTime);
		}
		
		//
		// Volume Control
		//

		private function stageMouseUp(e:MouseEvent):void
		{
			_volumeMouseDown = false;
			_volumeSlider.stopDrag();

			if (e.target != _volumePanel && !_volumePanel.contains(DisplayObject(e.target)))
			{
				// onReleaseOutside
				closeVolumePanel();
			}
			stage.removeEventListener(MouseEvent.MOUSE_UP, stageMouseUp);
		}
		
		private function showVolumePanel(e:MouseEvent):void
		{
			_volumePanel.visible = true;
		}

		private function hideVolumePanel(e:MouseEvent):void
		{
			_volumePanel.visible = false;
		}
		
		private function toggleVolumePanel(e:MouseEvent):void
		{
			_volumePanel.visible = !_volumePanel.visible;
		}

		private function volumePanelRollOver(e:MouseEvent):void
		{
			_volumeMouseOver = true;
		}

		private function volumePanelRollOut(e:MouseEvent):void
		{
			_volumeMouseOver = false;
			setTimeout(closeVolumePanel,500);
		}

		private function volumeSliderMouseDown(e:MouseEvent):void
		{
			_volumeMouseDown = true;
			_volumeSlider.startDrag(true, new Rectangle(_volumeSlider.x, 3, 0, _volumePost.height));
			stage.addEventListener(MouseEvent.MOUSE_UP, stageMouseUp);
		}

		private function volumeSliderMouseMove(e:MouseEvent):void
		{
			_mute   = false;
			_volume = getVolumeFromSlider();
			this.setVolume();
		}

		private function closeVolumePanel():void
		{
			if (_volumePanel.visible && !_volumeMouseDown && !_volumeMouseOver)
			{
				_volumeSlider.stopDrag();
				
				_mute   = false;
				_volume = getVolumeFromSlider();
				this.setVolume();

//				TransitionManager.start(_volumePanel, {type:Fade, direction:Transition.OUT, duration:.3, easing:None.easeNone, startPoint:1});
				TransitionManager.start(_volumePanel, {type:Fly, direction:Transition.OUT, duration:.4, easing:None.easeNone, startPoint:8});
			}
		}

		private function volumePostClick(e:MouseEvent):void
		{
			setVolumeSlider(1 - _volumePost.mouseY/_volumePost.height);
			
			_mute   = false;
			_volume = getVolumeFromSlider();
			this.setVolume();
		}

		private function setVolumeSlider(v:Number):void
		{
			_volumeSlider.y = _volumePost.y + (1 - v) * _volumePost.height - _volumeSlider.height/2;	
		}

		private function getVolumeFromSlider():Number
		{
			return 1 - (_volumeSlider.y - _volumePost.y + _volumeSlider.height/2) / _volumePost.height;
		}
	}
}
