﻿/* 
 *  Author:  Flavio Ribeiro (fr@lps.usp.br)
 *  Date:    January 15, 2011
 *  Version: 0.1
 *
 *  Copyright (c) 2011 Flavio Ribeiro
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.  
 */

package 
{	
	import flash.display.*;
	import flash.events.*;
	import flash.filters.*;
	import flash.net.*;
	import flash.text.*;
	import flash.utils.*;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Matrix;
	import flash.ui.Mouse;
	
	import com.hurlant.util.*;
	import utils.sprintf;

	public class MTurkCalibration extends FCalibrationPlayer
	{
		private static const ALPHA:Number = 0.8;	// low pass filter coefficient for the accuracy needle
		private static const HZ:int = 25;			// number of mouseHistory captures per second
		private static const DELTAT:Number = 0.25;	// time shift when replaying the mouse history
		
		private var _blurContainer:Sprite = new Sprite();
		private var _blurredFrame:Bitmap = new Bitmap();
		private var _alphaMask:Bitmap = new Bitmap();
		private var _colorMask:Bitmap = new Bitmap();

		[Embed(source="pbkernels/gradient_alpha_norm2.pbj", mimeType="application/octet-stream")]
		private var _alphaGradientKernel:Class;
		private var _alphaGradientShader:Shader = new Shader(new _alphaGradientKernel());
		private var _alphaGradientFilter:ShaderFilter = new ShaderFilter(_alphaGradientShader);

		[Embed(source="pbkernels/gradient_luminance_norm2.pbj", mimeType="application/octet-stream")]
		private var _colorGradientKernel:Class;
		private var _colorGradientShader:Shader = new Shader(new _colorGradientKernel());
		private var _colorGradientFilter:ShaderFilter = new ShaderFilter(_colorGradientShader);

		private var _answerFrame:Shape = new Shape();
		private var _fpsText:TextField = new TextField();
		private var _userMessage:TextField = new TextField();
		private var _speedometer:speedometer = new speedometer();
		private var _speedometerNeedle:speedometerNeedle = new speedometerNeedle();
		private var _legend:legend = new legend();
		private var _crosshair:crosshair = new crosshair();
		private var _bigSubmitButton:bigSubmitButton = new bigSubmitButton();

		private var _accuracy:Number = 0;
		private var _startTime:Number = 0;
		private var _prevFrameTime:Number;
		private var _frameCounter:uint = 0;
		private var _mouseCounter:uint = 0;

		private var _submissionCount:int = 0;			// counter of submission attempts (-1 when successfully submitted)
		private var _validAssignment:Boolean = true;	// assignment becomes invalid if the video stops unexpectedly (for example, due to buffering)
		private var _mouseHistory:MouseHistory = new MouseHistory();
		private var _kmeanHistory:Array = new Array();
		private var _answerMode:Boolean = false;

		public function MTurkCalibration()
		{
			this.parseArguments();
			this.attachWidgets();
			this.drawWidgets();
			this.addEventCallbacks();
		}

		private function isPreview():Boolean
		{
			return !_queryString.parameters.assignmentId || _queryString.parameters.assignmentId == "ASSIGNMENT_ID_NOT_AVAILABLE";
		}

		private function isReview():Boolean
		{
			return _queryString.parameters.reviewId;
		}

		protected override function processConfigXML(e:Event):void
		{
			var coordX:Array;
			var coordY:Array;
			var timestamp:Array
			
			super.processConfigXML(e);

			try
			{
				if (isReview())
				{
					var reviewId:String = _queryString.parameters.reviewId;

					coordX    = Base64.decodeDecompress(_configXML.video.(@id == _videoId).solution.(@id == reviewId).coordX).split(',');
					coordY    = Base64.decodeDecompress(_configXML.video.(@id == _videoId).solution.(@id == reviewId).coordY).split(',');
					timestamp = Base64.decodeDecompress(_configXML.video.(@id == _videoId).solution.(@id == reviewId).timestamp).split(',');

					if (coordX.length <= 1 || coordY.length <= 1 || timestamp.length <= 1)
						throw new Error();

					_mouseHistory = new MouseHistory();
					_mouseHistory.init(coordX, coordY, timestamp);
				}
			}
			catch(e:Error)
			{
				_mouseHistory = null;
			}
			
			try
			{
				var k:int = 0;
				for each (var solution in _configXML.video.(@id == _videoId).solution)
				{
					coordX    = Base64.decodeDecompress(solution.coordX).split(',');
					coordY    = Base64.decodeDecompress(solution.coordY).split(',');
					timestamp = Base64.decodeDecompress(solution.timestamp).split(',');
					
					if (coordX.length <= 1 || coordY.length <= 1 || timestamp.length <= 1)
						throw new Error();

					_kmeanHistory[k] = new MouseHistory();
					_kmeanHistory[k].init(coordX, coordY, timestamp);
					k++;
				}
				
				if (k == 0)
				{
					_kmeanHistory  = null;
					_replayEnabled = false;
				}
			}
			catch(e:Error)
			{
				_kmeanHistory  = null;
				_replayEnabled = false;
			}
		}
/*
		private function submitHostAssignmentComplete(e:Event):void
		{
			_submissionCount = -1;
		}

		private function submitHostAssignmentError(e:IOErrorEvent):void
		{
			if (_submissionCount < 5)
				submitHostAssignment();
			else
				_submissionCount = -1;
		}

		private function submitHostAssignment():void
		{
			if (!isReview() && !isPreview() && _submissionCount != -1)
			{
				var variables:URLVariables = new URLVariables();
				var host_request:URLRequest = new URLRequest();
				var host_loader:URLLoader = new URLLoader();

				variables.hitId        = _queryString.parameters.hitId;
				variables.assignmentId = _queryString.parameters.assignmentId;
				variables.videoId      = _queryString.parameters.videoId;
				variables.valid        = _validAssignment ? 'true' : 'false';

				// compress with zlib and Base64 encode the assignment (which will be POSTed to the MTurk server)
				variables.coordX       = Base64.compressEncode(_mouseHistory.getCoordXVector().join(","));
				variables.coordY       = Base64.compressEncode(_mouseHistory.getCoordYVector().join(","));
				variables.timestamp    = Base64.compressEncode(_mouseHistory.getTimeStampVector().join(","));

				host_request.url = "store_assignment.php";
				host_request.method = URLRequestMethod.POST;
				host_request.data = variables;

				host_loader.addEventListener(Event.COMPLETE, submitHostAssignmentComplete);
				host_loader.addEventListener(IOErrorEvent.IO_ERROR, submitHostAssignmentError);

				++_submissionCount;
				host_loader.load(host_request);
//				navigateToURL(host_request, "_self");
			}
		}
*/
		private function submitTurkAssignment(e:MouseEvent):void
		{
			if (!isReview() && !isPreview())
			{
				var variables:URLVariables = new URLVariables();
				var turk_request:URLRequest = new URLRequest();
				var turk_loader:URLLoader = new URLLoader();

				variables.hitId        = _queryString.parameters.hitId;
				variables.assignmentId = _queryString.parameters.assignmentId;
				variables.videoId      = _queryString.parameters.videoId;
				variables.valid        = _validAssignment ? 'true' : 'false';

				// compress with zlib and Base64 encode the assignment (which will be POSTed to the MTurk server)
				variables.coordX       = Base64.compressEncode(_mouseHistory.getCoordXVector().join(","));
				variables.coordY       = Base64.compressEncode(_mouseHistory.getCoordYVector().join(","));
				variables.timestamp    = Base64.compressEncode(_mouseHistory.getTimeStampVector().join(","));
	
				turk_request.url = (_queryString.url.indexOf('workersandbox') != -1) ? "https://workersandbox.mturk.com/mturk/externalSubmit" : "https://www.mturk.com/mturk/externalSubmit";
				turk_request.method = URLRequestMethod.POST;
				turk_request.data = variables;

				navigateToURL(turk_request, "_self");
			}
		}

		private function parseArguments():void
		{
			var parameters = LoaderInfo(this.root.loaderInfo).parameters;

			if (isPreview())
			{
				_replayEnabled = false;
				setImageURL("preview.png");
			}

			if (isReview())
				_answerMode = true;				
		}

		private function attachWidgets():void
		{
			_speedometer.addChild(_speedometerNeedle);
			_blurContainer.addChild(_blurredFrame);
			_blurContainer.addChild(_alphaMask);
			
			_videoOverlay.addChild(_blurContainer);
			_videoOverlay.addChild(_colorMask);

			_videoOverlay.addChild(_answerFrame);
//			_videoOverlay.addChild(_fpsText);
			_videoOverlay.addChild(_speedometer);
			_videoOverlay.addChild(_legend);
			_videoOverlay.addChild(_crosshair);

			_videoOverlay.addChild(_userMessage);
			_videoOverlay.addChild(_bigSubmitButton);
		}
		
		// accuracy: number between 0 and 100 representing user accuracy
		private function rotateNeedle(accuracy:Number):void
		{
			var angle:Number = accuracy/100 * 250;
			var needleMatrix:Matrix = _speedometerNeedle.transform.matrix;
			needleMatrix.identity();
			needleMatrix.translate(39.7/2, -39.7/2);
			needleMatrix.rotate(angle/180.0 * Math.PI);
			_speedometerNeedle.transform.matrix = needleMatrix;
		}
		
		private function updateUserMessage(message:String):void
		{
			_userMessage.text = message;
		}

		private function drawWidgets():void
		{
			var videoWidth = _canvas.width;
			var videoHeight = _canvas.height;
			var alphaMaskBitmapData = new BitmapData(2*videoWidth, 2*videoHeight, true, 0x00000000);
			var colorMaskBitmapData = new BitmapData(2*videoWidth, 2*videoHeight, true, 0x00000000);
			var maskRectangle = new Rectangle(0, 0, 2*videoWidth, 2*videoHeight);
			var maskPoint = new Point(0, 0);
			
 			_alphaGradientShader.data.radius.value = [ 30 ];
			_alphaGradientShader.data.center.value = [ videoWidth, videoHeight ];
 			_colorGradientShader.data.radius.value = [ 30 ];
			_colorGradientShader.data.center.value = [ videoWidth, videoHeight ];
			alphaMaskBitmapData.applyFilter(alphaMaskBitmapData, maskRectangle, maskPoint, _alphaGradientFilter);
			colorMaskBitmapData.applyFilter(colorMaskBitmapData, maskRectangle, maskPoint, _colorGradientFilter);

			_alphaMask.bitmapData = alphaMaskBitmapData;
//			_colorMask.bitmapData = colorMaskBitmapData;
			_alphaMask.blendMode = BlendMode.ALPHA;
//			_colorMask.blendMode = BlendMode.MULTIPLY;
			_blurContainer.blendMode = BlendMode.LAYER;

			var _blurFilter:BitmapFilter;
			_blurFilter = new BlurFilter(10, 10, BitmapFilterQuality.LOW);
			_blurredFrame.bitmapData = new BitmapData(videoWidth, videoHeight, true, 0x00000000);
			_blurredFrame.filters = [ _blurFilter ];

			_speedometer.scaleX = _speedometer.scaleY = .4;
			_speedometer.x = stage.stageWidth - _speedometer.width/2 - 4;
			_speedometer.y = _speedometer.height/2 + 4;
			_speedometer.cacheAsBitmap = true;

			_speedometerNeedle.transform.matrix = new Matrix();
			_speedometerNeedle.cacheAsBitmap = true;
			rotateNeedle(0);

			_legend.scaleX = _legend.scaleY = .4;
			_legend.x = _speedometer.x - _speedometer.width/2;
			_legend.y = _speedometer.y + _speedometer.width/2 + 4;
			_legend.cacheAsBitmap = true;
			
			_crosshair.visible = false;
			_crosshair.cacheAsBitmap = true;
			
			_fpsText.defaultTextFormat = new TextFormat("_sans", 14, 0xffffff);
			
			_userMessage.defaultTextFormat = new TextFormat("_sans", 16, 0xffffff);
			_userMessage.width = videoWidth;
			_userMessage.x = 4;
			_userMessage.y = videoHeight - 20 - 4;
			
			_bigSubmitButton.x = videoWidth  - _bigSubmitButton.width  - 4;
			_bigSubmitButton.y = videoHeight - _bigSubmitButton.height - 4;
			_bigSubmitButton.visible = false;

			this.renderState(_state);
		}

		public override function renderState(state:String):void
		{
			super.renderState(state);

			_speedometer.visible = false;
			_legend.visible = false;
			_answerFrame.visible = false;

			switch(state)
			{
				case STOPPED:
					break;

				case PLAYING:
					_mouseCounter = 0;
					_startTime = getTimer();

				case PAUSED:
					if (isReview())
						updateUserMessage('Running in review mode. Results won\'t be submitted.');
					else if (isPreview())
						updateUserMessage('This is only a preview. Please ACCEPT the HIT to work on this task.');
					else
						updateUserMessage('');
					
					_speedometer.visible = _mouseHistory && _kmeanHistory;
					_legend.visible      = _mouseHistory && _kmeanHistory && _answerMode;
					_answerFrame.visible = _answerMode;
					
					_blurContainer.visible   = !_answerMode;
					_colorMask.visible       = !_answerMode;
					_crosshair.visible       = !_answerMode;
					break;

				case COMPLETE:
					_blurContainer.visible   = false;
					_colorMask.visible       = false;
					_crosshair.visible       = false;
					_bigSubmitButton.visible = !isPreview() && !isReview();
					break;

				case ERROR:
					break;

				default:
					break;
			}
		}

		private function addEventCallbacks():void
		{
			// moveMask must not be called too often, because it is CPU intensive.
			// saveMouse and drawAnswer are called using the timer event, which allows for faster and more accurate timing
			// (the playhead isn't always updated between two EnterFrame events, and small values of playheadUpdateInterval make the video status jump to COMPLETE on the Linux player)

			var frameTimer:Timer = new Timer(1/HZ * 1000);
			frameTimer.addEventListener(TimerEvent.TIMER, timerEventHandler);	// capture pointer and draw answers
			frameTimer.start();

			addEventListener(Event.ENTER_FRAME, enterFrameHandler);				// update FPS, show/hide pointer and move blur mask
			_bigSubmitButton.addEventListener(MouseEvent.CLICK, submitTurkAssignment);
		}

		//
		// Event Handlers
		//

		private function enterFrameHandler(e:Event)
		{
			var videoWidth  = _canvas.width;
			var videoHeight = _canvas.height;
			var mouseX = Math.min(_canvas.mouseX, videoWidth);
			var mouseY = Math.min(_canvas.mouseY, videoHeight);
			
			// update FPS
			if (_frameCounter++ % 20 == 0)
			{
				var now:Number = getTimer();
				if (!isNaN(_prevFrameTime))
					_fpsText.text = "FPS " + Math.round(1000 / (now - _prevFrameTime) * 20);
				_prevFrameTime = now;
			}
			
			// show/hide mouse pointer
			if (_state == PLAYING && !_answerMode && mouseX < videoWidth && mouseY < videoHeight)
			{
				_crosshair.x = mouseX;
				_crosshair.y = mouseY;
				Mouse.hide();
			}
			else
				Mouse.show();

			// update blur mask
			if (_state == PLAYING && !_answerMode)
			{
				_colorMask.x = -videoWidth  + mouseX;
				_colorMask.y = -videoHeight + mouseY;
				
				_alphaMask.x = -videoWidth  + mouseX;
				_alphaMask.y = -videoHeight + mouseY;

				_blurredFrame.bitmapData.draw(_canvas);
			}
		}
		
		private function norm(x:Number, y:Number):Number
		{
			return Math.sqrt(x*x + y*y);
		}

		private function timerEventHandler(e:Event)
		{
			if (_state == PLAYING)
			{
				var timeStamp:Number = (getTimer() - _startTime)/1000;
				var minDist:Number = 1e10;
				
				var videoWidth :Number = _canvas.width;
				var videoHeight:Number = _canvas.height;
				var mouseX:Number = Math.min(_canvas.mouseX, videoWidth);
				var mouseY:Number = Math.min(_canvas.mouseY, videoHeight);
				var coordX:Number;
				var coordY:Number;
				var k:int;
				var K:int;

				if (_answerMode)
				{
					// in answer mode -- draw mouse pointer locations and K-means clusters
					var radius:Number = 80;
					var matrix:Matrix = new Matrix();

					_answerFrame.graphics.clear();
					
					K = _kmeanHistory ? _kmeanHistory.length : 0;
					for (k = 0; k < K; k++)
					{						
						_kmeanHistory[k].seekToTimeStamp(timeStamp + DELTAT);
						coordX = _kmeanHistory[k].getCoordX();
						coordY = _kmeanHistory[k].getCoordY();

						if (coordX >= 0 && coordY >= 0)
						{
							coordX += (_canvas.width  - 640)/2;
							coordY += (_canvas.height - 360)/2;

							minDist = Math.min(minDist, norm(mouseX - coordX, mouseY - coordY));	
							matrix.createGradientBox(2*radius, 2*radius, 0, coordX - radius, coordY - radius);
							_answerFrame.graphics.beginGradientFill(GradientType.RADIAL, [0x0000FF, 0x0000FF], [.00, .40], [192, 255], matrix);
							_answerFrame.graphics.drawCircle(coordX, coordY, radius);
							_answerFrame.graphics.endFill();
						}
					}

					if (_mouseHistory)
					{
						_mouseHistory.seekToTimeStamp(timeStamp + DELTAT);
						mouseX = _mouseHistory.getCoordX();
						mouseY = _mouseHistory.getCoordY();

						if (mouseX >= 0 && mouseY >= 0)
						{
							mouseX += (_canvas.width  - 640)/2;
							mouseY += (_canvas.height - 360)/2;

							matrix.createGradientBox(2*radius, 2*radius, 0, mouseX - radius, mouseY - radius);
							_answerFrame.graphics.beginGradientFill(GradientType.RADIAL, [0xFF0000, 0xFF0000], [.00, .40], [192, 255], matrix);
							_answerFrame.graphics.drawCircle(mouseX, mouseY, radius);
							_answerFrame.graphics.endFill();
						}
					}
				}
				else
				{
					// not in answer mode -- store mouse pointer coordinates
					mouseX -= (_canvas.width  - 640)/2;
					mouseY -= (_canvas.height - 360)/2;

					_mouseCounter++;
					_mouseHistory.push(mouseX, mouseY, timeStamp);
//					_mouseHistory.push(_ballX, _ballY, timeStamp);

					if (timeStamp > 10 && _mouseCounter < 0.95 * timeStamp * HZ)
					{
						printError("<p align=\"center\">Error: your computer is too slow to work on this HIT. Sorry :(</p>");
						pauseVideo(null);
					}

					K = _kmeanHistory ? _kmeanHistory.length : 0;
					for (k = 0; k < K; k++)
					{
						_kmeanHistory[k].seekToTimeStamp(timeStamp);
						coordX = _kmeanHistory[k].getCoordX();
						coordY = _kmeanHistory[k].getCoordY();
						
						if (coordX >= 0 && coordY >= 0)
							minDist = Math.min(minDist, norm(mouseX - coordX, mouseY - coordY));
					}
				}
				
				_accuracy = ALPHA * _accuracy + (1 - ALPHA) * Math.max(100*(1 - 4*minDist/videoWidth), 0.0);
				rotateNeedle(_accuracy);
			}
		}

		public override function videoComplete():void
		{
//			submitHostAssignment();
			super.videoComplete();
		}

		public override function replayVideo(e:MouseEvent)
		{
			_answerMode = true;
			super.replayVideo(e);
		}
	}
}
