﻿/* 
 *  Author:  Flavio Ribeiro (fr@lps.usp.br)
 *  Date:    January 15, 2011
 *  Version: 0.1
 *
 *  Copyright (c) 2011 Flavio Ribeiro
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.  
 */

package
{
	public class MouseHistory
	{
		private static const INIT_SIZE:int = 1024;
		private var _coordX:Array      = new Array(INIT_SIZE);
		private var _coordY:Array      = new Array(INIT_SIZE);
		private var _timeStamp:Array   = new Array(INIT_SIZE);
		private var _usedSize:int      = 0;
		private var _allocatedSize:int = INIT_SIZE;
		private var _lastReadIndex:int = 0;

		private static const ALPHA:Number = 0.5;
		private var _lastX:int = 0;
		private var _lastY:int = 0;
		
		public function MouseHistory()
		{
		}

		public function init(coordX:Array, coordY:Array, timeStamp:Array)
		{
			if (coordX.length == coordY.length &&
				coordX.length == timeStamp.length &&
				coordY.length == timeStamp.length)
			{
				_coordX    = coordX;
				_coordY    = coordY;
				_timeStamp = timeStamp;
				
				_usedSize      = coordX.length;
				_allocatedSize = coordX.length;
				_lastReadIndex = 0;
			}
		}

		public function seekToTimeStamp(timeStamp:Number):void
		{
			var last_index:int = _usedSize - 1;
			
			while (_lastReadIndex < last_index &&
				   Math.abs(_timeStamp[_lastReadIndex+1] - timeStamp) <= Math.abs(_timeStamp[_lastReadIndex] - timeStamp))
				_lastReadIndex++;

			while (_lastReadIndex > 0 &&
				   Math.abs(_timeStamp[_lastReadIndex-1] - timeStamp) <= Math.abs(_timeStamp[_lastReadIndex] - timeStamp))
				_lastReadIndex--;
		}
		
		// get a low-pass filtered version of the X coordinate
		public function getCoordX():Number
		{
			if (_coordX[_lastReadIndex] < 0)
				return _coordX[_lastReadIndex];
			else
			{
				_lastX = ALPHA * _lastX + (1-ALPHA) * _coordX[_lastReadIndex];
				return _lastX;
			}
		}
		
		// get a low-pass filtered version of the Y coordinate
		public function getCoordY():Number
		{
			if (_coordY[_lastReadIndex] < 0)
				return _coordY[_lastReadIndex];
			else
			{
				_lastY = ALPHA * _lastY + (1-ALPHA) * _coordY[_lastReadIndex];
				return _lastY;
			}
		}
		
		public function getTimeStamp():Number
		{
			return _timeStamp[_lastReadIndex];
		}

		public function push(coordX:Number, coordY:Number, timeStamp:Number):void
		{
			if (_usedSize == _allocatedSize)
			{
				_allocatedSize *= 2;
				_coordX[_allocatedSize]    = null;
				_coordY[_allocatedSize]    = null;
				_timeStamp[_allocatedSize] = null;
			}

			_coordX[_usedSize]    = coordX;
			_coordY[_usedSize]    = coordY;
			_timeStamp[_usedSize] = timeStamp;
			_usedSize++;
		}
		
		public function getCoordXVector():Array
		{
			return _coordX.slice(0,_usedSize-1);
		}
		
		public function getCoordYVector():Array
		{
			return _coordY.slice(0,_usedSize-1);
		}
		
		public function getTimeStampVector():Array
		{
			return _timeStamp.slice(0,_usedSize-1);
		}
	}
}
